VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsCam"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
'============================================================
'== Camera class (clsCam) Test project    by Mike Brascome ==
'== Visit collective.freeservers.com/visbasic.htm          ==
'============================================================
'Properties: yaw                - Current yaw angle     (if MonitorAngles = True )
'            pitch              - Current pitch angle   (if MonitorAngles = True )
'            roll               - Current roll angle    (if MonitorAngles = True )
'            MonitorAngles      - Determines if the class monitors the Euler angles.
'                                 If this is set to False, this class will not update
'                                 yaw, pitch, and roll.
'            RotationMatrix     - Matrix containing rotation data
'            TranslationMatrix  - Matrix containing translation data
'            ViewMatrix         - Matrix containing both rotation and translation data
'            Viewport           - Contains the viewport we are rendering to
'            ProjectionMatrix   - Contains the projection matrix we are rendering with
'            Position           - vector indicating coordinates of the camera
'            Top                - vector indicating which direction is up
'            Front              - vector indicating which direction the camera is facing
'            Right              - vector indicating which direction is the camera's right
'
'Methods: InitCamAng    - Initializes the camera using position and angles
'         InitCamVec    - Initializes the camera using position and lookat vectors
'         InitCommon    - Initialization common to InitCamAng and InitCamVec
'         InView        - Returns true if the specified 3d object is in view.  Also returns
'                         2d screen coordinates and distance
'         RotateAbs     - Rotate to an absolute angle
'         RotateRel     - Rotate relative to current position
'         MoveAbs       - Move to absolute location
'         MoveRel       - Move relative to current position
'         MoveForward   - Move in the direction the camera is currently facing
'         LookAt        - Points the camera at a position in 3D space
'         LookAtSetTop  - Points the camera at a position in 3D space and sets the up direction
'
'Events: CamError(Msg)  - This event fires if the camera encounters an internal error.
'
'==============================================================================
'USAGE: The camera object will be declared like this:
'    Dim CamObj as new clsCam
'
'    It can then be initialized like this:
'    Call CamObj.InitCamAng(veiwport, projection, vEyePt, yaw, pitch, roll)
'
'    Or, if it can be init. with vectors like this:
'    Call CamObj.InitCamVec(veiwport, projection, vEyePt, vLookAtpt, vUp)
'
'    Once it's been set up, it can be controled like this:
'
'    'Yaw
'    If KeyLeft Then Call CamObj.RotateRel(1, 0, 0)
'    If KeyRight Then Call CamObj.RotateRel(-1, 0, 0)
'    'Pitch
'    If KeyUp Then Call CamObj.RotateRel(0, 1, 0)
'    If KeyDown Then Call CamObj.RotateRel(0, -1, 0)
'    'Roll
'    If RollLeft Then Call CamObj.RotateRel(0, 0, 1)
'    If RollRight Then Call CamObj.RotateRel(0, 0, -1)
'    'Straight up and down
'    If Yup Then Call CamObj.MoveRel(0, 1, 0)
'    If Ydown Then Call CamObj.MoveRel(0, -1, 0)
'    'forward and backward
'    If bSpace Then Call CamObj.MoveForward(1)
'    If BackUp Then Call CamObj.MoveForward(-1)
'    g_dev.SetTransform D3DTS_VIEW, CamObj.ViewMatrix
'
'    The camera can be moved around freely in 3d space very simply in just 11 lines
'    lines of code.
'
'    Here is a sample of pulling a 3d objects 2d screen coordinates for a heads-up display
'    If CamObj.InView(x, y, z, sngDistance, d3dvecPostion) Then
'        TextRect.Top = y
'        TextRect.Left = x
'        TextRect.bottom = y + 16
'        TextRect.Right = x + 200
'        D3DX.DrawText GameFont, &HFFFFFFFF, strName(i), TextRect, DT_TOP Or DT_LEFT
'        'Find out which object is closest
'        If sngDistance < sngClosest Then
'            sngClosest = sngDistance
'            strClsName = strName(i)
'        End If
'    Else
'        'This tells Render that the planet is not in view
'        UpdatePlanet.z = 999
'    End If

Option Explicit

Private Const PI = 3.14                     'Constants
Private Const RAD = PI / 180
Private Const DEG = 180 / PI

Private vTop As D3DVECTOR                   'vector indicating which direction is up
Private vFront As D3DVECTOR                 'vector indicating which direction the camera is facing
Private vRight As D3DVECTOR                 'vector indicating the camera's right

Private mvarTranslationMatrix As D3DMATRIX  'Matrix containing the camera's postion
Private mvarRotationMatrix As D3DMATRIX     'Matrix containing the camera's rotation
Private mvarYaw As Single                   'Euler rotations.
Private mvarPitch As Single                 'We keep up with these in case an external app
Private mvarRoll As Single                  'needs them.  This class doesn't need them.

'The following are used to calculate vTop, vFront, and vRight
Private lViewport As D3DVIEWPORT8           'Contains the viewport we are rendering to
Private lProjection As D3DMATRIX            'Contains the projection matrix we are rendering with
Private lAspect As Single                   'Aspect ratio of the projection matrix
Private matWorld As D3DMATRIX               'Default world matrix
Private vSrcMid As D3DVECTOR                'Center of the veiwport
Private vSrcRht As D3DVECTOR                'Point at middle right in veiwport
Private vCntr As D3DVECTOR                  'Center of camera's default matrix

'This event fires if the camera encounters an internal error.
Public Event CamError(strMsg As String)

Public MonitorAngles As Boolean             'Determines if the class monitors the Euler angles

Public Sub InitCamAng(Povy As Single, aspect As Single, zn As Single, zf As Single, viewport As D3DVIEWPORT8, bMonitorAngles As Boolean, pos As D3DVECTOR, yaw As Single, pitch As Single, roll As Single)
    On Error GoTo CamErr
    Call InitCommon(Povy, aspect, zn, zf, viewport, bMonitorAngles)
    D3DXMatrixRotationYawPitchRoll mvarRotationMatrix, yaw * RAD, pitch * RAD, roll * RAD
    D3DXMatrixTranslation mvarTranslationMatrix, -pos.x, -pos.y, -pos.z
    Call CalcTopNFront
    'Record yaw, pitch, and roll
    mvarYaw = yaw * RAD
    mvarPitch = pitch * RAD
    mvarRoll = roll * RAD
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error initializing camera (" & Err.Number & ":" & Err.Description & ")")
End Sub
Public Sub InitCamVec(Povy As Single, aspect As Single, zn As Single, zf As Single, viewport As D3DVIEWPORT8, bMonitorAngles As Boolean, pos As D3DVECTOR, vLookAt As D3DVECTOR, vUp As D3DVECTOR)
    Dim matTempMatrix As D3DMATRIX
    Dim CamPos As D3DVECTOR
    On Error GoTo CamErr
    
    Call InitCommon(Povy, aspect, zn, zf, viewport, bMonitorAngles)     'Update common var
    Call MoveAbs(pos.x, pos.y, pos.z)                                   'Set translation matrix
    CamPos.x = -mvarTranslationMatrix.m41
    CamPos.y = -mvarTranslationMatrix.m42
    CamPos.z = -mvarTranslationMatrix.m43
        
    D3DXMatrixLookAtLH mvarRotationMatrix, CamPos, vLookAt, vUp
    'Send the rotation matrix back to the origin
    mvarRotationMatrix.m41 = 0
    mvarRotationMatrix.m42 = 0
    mvarRotationMatrix.m43 = 0

    Call CalcTopNFront
    'Record yaw, pitch, and roll
    Call ExtractYawPitchRoll(pos, vLookAt, vUp)
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error initializing camera (" & Err.Number & ":" & Err.Description & ")")
    
End Sub
Private Sub InitCommon(Povy As Single, aspect As Single, zn As Single, zf As Single, viewport As D3DVIEWPORT8, bMonitorAngles As Boolean)
    On Error GoTo CamErr
    
    MonitorAngles = bMonitorAngles
    If MonitorAngles = False Then
        mvarYaw = 0
        mvarPitch = 0
        mvarRoll = 0
    End If
    D3DXMatrixIdentity mvarTranslationMatrix
    D3DXMatrixIdentity mvarRotationMatrix
    
    lViewport = viewport
    ' Define the projection matrix.
    D3DXMatrixPerspectiveFovLH lProjection, Povy, aspect, zn, zf
    lAspect = aspect
    
    D3DXMatrixIdentity matWorld
    vSrcMid.x = lViewport.Width / 2
    vSrcMid.y = lViewport.Height / 2
    vSrcMid.z = 0
    vSrcRht.x = lViewport.Width
    vSrcRht.y = lViewport.Height / 2
    vSrcRht.z = 0
    vCntr.x = 0
    vCntr.y = 0
    vCntr.z = 0
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error initializing camera (" & Err.Number & ":" & Err.Description & ")")
End Sub
' ======================= Methods ========================
Public Sub RotateAbs(ByVal yaw As Single, ByVal pitch As Single, ByVal roll As Single)
    On Error GoTo CamErr
    D3DXMatrixRotationYawPitchRoll mvarRotationMatrix, yaw * RAD, pitch * RAD, roll * RAD
    Call CalcTopNFront
    If MonitorAngles Then
        'Record yaw, pitch, and roll
        mvarYaw = yaw * RAD
        mvarPitch = pitch * RAD
        mvarRoll = roll * RAD
    End If
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in RotateAbs (" & Err.Number & ":" & Err.Description & ")")
End Sub

Public Sub RotateRel(ByVal yaw As Single, ByVal pitch As Single, ByVal roll As Single)
    Dim tmpRotMatrix As D3DMATRIX
    
    On Error GoTo CamErr
    'Create an update matrix
    D3DXMatrixIdentity tmpRotMatrix
    D3DXMatrixRotationYawPitchRoll tmpRotMatrix, yaw * RAD, pitch * RAD, roll * RAD
    
    'Add it to the rot matrix
    D3DXMatrixMultiply mvarRotationMatrix, mvarRotationMatrix, tmpRotMatrix
    If MonitorAngles Then
        mvarYaw = mvarYaw + yaw * RAD
        mvarPitch = mvarPitch + pitch * RAD
        mvarRoll = mvarRoll + roll * RAD
        If mvarYaw > PI Then mvarYaw = mvarYaw - 2 * PI
        If mvarYaw < -PI Then mvarYaw = mvarYaw + 2 * PI
        If mvarPitch > PI Then mvarPitch = mvarPitch - 2 * PI
        If mvarPitch < -PI Then mvarPitch = mvarPitch + 2 * PI
        If mvarRoll > PI Then mvarRoll = mvarRoll - 2 * PI
        If mvarRoll < -PI Then mvarRoll = mvarRoll + 2 * PI
    End If
    Call CalcTopNFront
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in RotateRel (" & Err.Number & ":" & Err.Description & ")")
End Sub

Public Sub MoveRel(ByVal x As Single, ByVal y As Single, ByVal z As Single)
    Dim tmpTrnsMatrix  As D3DMATRIX
    
    On Error GoTo CamErr
    'Create an update matrix
    D3DXMatrixIdentity tmpTrnsMatrix
    D3DXMatrixTranslation tmpTrnsMatrix, x, y, z
    'Add it to the trans matrix
    D3DXMatrixMultiply mvarTranslationMatrix, mvarTranslationMatrix, tmpTrnsMatrix
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in MoveRel (" & Err.Number & ":" & Err.Description & ")")
End Sub

Public Sub MoveAbs(ByVal x As Single, ByVal y As Single, ByVal z As Single)
    On Error GoTo CamErr
    D3DXMatrixTranslation mvarTranslationMatrix, -x, -y, -z
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in MoveAbs (" & Err.Number & ":" & Err.Description & ")")
End Sub

Public Sub MoveForward(ByVal Dist As Single)
    Dim tmpTrnsMatrix  As D3DMATRIX
    
    On Error GoTo CamErr
    'Create a vector in front of the camera Dist units long
    D3DXMatrixIdentity tmpTrnsMatrix
    D3DXMatrixTranslation tmpTrnsMatrix, vFront.x * -Dist, vFront.y * -Dist, vFront.z * -Dist
    'Move the camera along that vector
    D3DXMatrixMultiply mvarTranslationMatrix, mvarTranslationMatrix, tmpTrnsMatrix
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in MoveFoward (" & Err.Number & ":" & Err.Description & ")")
End Sub

Public Sub LookAtSetTop(vLookAt As D3DVECTOR, vUp As D3DVECTOR)
    Dim CamPos As D3DVECTOR
    
    On Error GoTo CamErr
    CamPos.x = mvarTranslationMatrix.m41
    CamPos.y = mvarTranslationMatrix.m42
    CamPos.z = mvarTranslationMatrix.m43
        
    D3DXMatrixLookAtLH mvarRotationMatrix, vec3(-CamPos.x, -CamPos.y, -CamPos.z), vLookAt, vUp
    'Send the rotation matrix back to the origin
    mvarRotationMatrix.m41 = 0
    mvarRotationMatrix.m42 = 0
    mvarRotationMatrix.m43 = 0
    
    Call CalcTopNFront
    'Record yaw, pitch, and roll
    If MonitorAngles Then Call ExtractYawPitchRoll(Position, vLookAt, vUp)
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in LookAt (" & Err.Number & ":" & Err.Description & ")")
    
End Sub
Public Sub LookAt(vLookAt As D3DVECTOR)
    Dim CamPos As D3DVECTOR
    
    On Error GoTo CamErr
    CamPos.x = mvarTranslationMatrix.m41
    CamPos.y = mvarTranslationMatrix.m42
    CamPos.z = mvarTranslationMatrix.m43
        
    D3DXMatrixLookAtLH mvarRotationMatrix, vec3(-CamPos.x, -CamPos.y, -CamPos.z), vLookAt, vTop
    'Send the rotation matrix back to the origin
    mvarRotationMatrix.m41 = 0
    mvarRotationMatrix.m42 = 0
    mvarRotationMatrix.m43 = 0
    
    Call CalcTopNFront
    'Record yaw, pitch, and roll
    If MonitorAngles Then Call ExtractYawPitchRoll(Position, vLookAt, vTop)
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in LookAt (" & Err.Number & ":" & Err.Description & ")")
    
End Sub
Public Function InView(ByRef vSX As Single, ByRef vSY As Single, ByRef vSZ As Single, ByRef DistToTarg As Single, vTarget As D3DVECTOR) As Boolean
        Dim lngOkay As Long
        Dim sngDot As Single
        Dim sngDist As Single
        Dim vCamVec As D3DVECTOR
        Dim VTargVec As D3DVECTOR
        
        D3DXVec3Subtract vCamVec, Position, vec3(Position.x + vFront.x, Position.y + vFront.y, Position.z + vFront.z)
        D3DXVec3Subtract VTargVec, Position, vTarget
        sngDist = D3DXVec3Length(VTargVec)
        D3DXVec3Normalize vCamVec, vCamVec
        D3DXVec3Normalize VTargVec, VTargVec
        sngDot = D3DXVec3Dot(vCamVec, VTargVec)
        DistToTarg = sngDist
        
        If sngDot > lAspect Then
            'target is in view
            
            'Calc screen coords
            Dim VOut As D3DVECTOR
            Dim V As D3DVECTOR
            Dim tmpWorld As D3DMATRIX
            
            D3DXVec3Subtract V, vTarget, Position
            D3DXVec3Normalize V, V
            D3DXMatrixIdentity tmpWorld
            tmpWorld.m41 = vTarget.x
            tmpWorld.m42 = vTarget.y
            tmpWorld.m43 = vTarget.z
                                      
            lngOkay = D3DXVec3Project(VOut, V, lViewport, lProjection, Viewmatrix, tmpWorld)
            If lngOkay <> D3DERR_INVALIDCALL And lngOkay <> D3DERR_OUTOFVIDEOMEMORY Then
                InView = True
                vSX = VOut.x
                vSY = VOut.y
                vSZ = VOut.z
            Else
                InView = False
            End If
        Else
            InView = False
        End If
'        Dim strMsg As String
'        strMsg = vbCrLf & "Mars  : " & Format(sngDot, "000.000")
'
'        frmSlrExp.lblStatus.Caption = frmSlrExp.lblStatus.Caption & vbCrLf & strMsg
End Function

' ====================== Properties ============================
Public Property Get yaw() As Single
    yaw = mvarYaw * DEG
End Property
Public Property Get pitch() As Single
    pitch = mvarPitch * DEG
End Property
Public Property Get roll() As Single
    roll = mvarRoll * DEG
End Property
Public Property Get RotationMatrix() As D3DMATRIX
    RotationMatrix = mvarRotationMatrix
End Property
Public Property Get TranslationMatrix() As D3DMATRIX
    TranslationMatrix = mvarTranslationMatrix
End Property
Public Property Get Viewmatrix() As D3DMATRIX
    Dim matView As D3DMATRIX
    D3DXMatrixMultiply matView, mvarTranslationMatrix, mvarRotationMatrix
    Viewmatrix = matView
End Property
Public Property Get viewport() As D3DVIEWPORT8
    viewport = lViewport
End Property
Public Property Get ProjectionMatrix() As D3DMATRIX
    ProjectionMatrix = lProjection
End Property
Public Property Get Top() As D3DVECTOR
    Top = vTop
End Property

Public Property Get Position() As D3DVECTOR
    Position.x = -mvarTranslationMatrix.m41
    Position.y = -mvarTranslationMatrix.m42
    Position.z = -mvarTranslationMatrix.m43
End Property
Public Property Get Front() As D3DVECTOR
    Front = vFront
End Property
Public Property Get Right() As D3DVECTOR
    Right = vRight
End Property
' =============================== Helper Functions ================================
Private Function vec3(ByVal x As Single, ByVal y As Single, ByVal z As Single) As D3DVECTOR
    vec3.x = x
    vec3.y = y
    vec3.z = z
End Function

Private Sub CalcTopNFront()
    On Error GoTo CamErr
    D3DXVec3Unproject vFront, vSrcMid, lViewport, lProjection, mvarRotationMatrix, matWorld
    'Get another point just to the right of vFront to complete a horizontal triangle(plane).
    D3DXVec3Unproject vRight, vSrcRht, lViewport, lProjection, mvarRotationMatrix, matWorld
    D3DXVec3Normalize vRight, vRight
    'Calculate the cross product of the horizontal triangle's first two sides will give us
    'it's normal which is vTop
    D3DXVec3Cross vTop, _
        vec3(vFront.x - vCntr.x, vFront.y - vCntr.y, vFront.z - vCntr.z), _
        vec3(vRight.x - vFront.x, vRight.y - vFront.y, vRight.z - vFront.z)
    D3DXVec3Normalize vTop, vTop
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in CalcTopNFront (" & Err.Number & ":" & Err.Description & ")")
End Sub

Private Sub ExtractYawPitchRoll(pos As D3DVECTOR, vLookAt As D3DVECTOR, vUp As D3DVECTOR)
    On Error GoTo CamErr
    'Figure out the angle the camera is at
    Dim vDir As D3DVECTOR
    D3DXVec3Subtract vDir, pos, vLookAt

    'yaw around y
    If vDir.x > 0 Then
        mvarYaw = Atn(vDir.z / vDir.x) + PI / 2
    ElseIf vDir.x < 0 Then
        mvarYaw = Atn(vDir.z / vDir.x) - PI / 2
    ElseIf vDir.x = 0 Then
        mvarYaw = 0
    End If

    'pitch around x
    If vDir.y > 0 Then
        mvarPitch = (Atn(Sqr(vDir.x ^ 2 + vDir.z ^ 2) / vDir.y)) - PI / 2
    ElseIf vDir.y < 0 Then
        mvarPitch = (Atn(Sqr(vDir.x ^ 2 + vDir.z ^ 2) / vDir.y)) + PI / 2
    ElseIf vDir.y = 0 Then
        mvarPitch = 0
    End If

    'roll around z
    'Still having problems calculating roll
    D3DXVec3Subtract vDir, pos, vec3(pos.x + vUp.x, pos.y + vUp.y, pos.z + vUp.z)
    If vDir.x > 0 Then
        mvarRoll = Atn(vDir.y / vDir.x)
    ElseIf vDir.x < 0 Then
        mvarRoll = Atn(vDir.y / vDir.x)
    ElseIf vDir.x = 0 Then
        mvarRoll = 0
    End If
    
    On Error GoTo 0
    Exit Sub
CamErr:
    RaiseEvent CamError("Error in ExtractYawPitchRoll (" & Err.Number & ":" & Err.Description & ")")
End Sub
